import json
import datetime
import secrets
from sqlalchemy.types import TypeDecorator, VARCHAR
from sqlalchemy.ext.mutable import Mutable
from sqlalchemy import (Column, Integer, Unicode, UnicodeText, DateTime,
    Boolean, ForeignKey, func)
from sqlalchemy.orm import relationship
from sqlalchemy.schema import UniqueConstraint
from tcw.database import Base


class JSONEncodedDict(TypeDecorator):
    """
    Store json text as a varchar
    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_variable(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value


class MutableDict(Mutable, dict):
    """
    Convert plain dictionaries to MutableDict
    """

    @classmethod
    def coerce(cls, key, value):
        """
        Coerce to dict
        """

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        """
        Detect dictionary set events and emit change events.
        """

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        """
        Detect dictionary del events and emit change events.
        """

        dict.__delitem__(self, key)
        self.changed()


class Contest(Base):
    """
    NOTE: Do not use the attributes field for ANYTHING.
    If you do, it will break the sqlalchemy subqueries that check for
    expired contests, and tcw-tasks will break.
    """

    __tablename__ = 'contests'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(48), nullable=False, unique=True)
    title = Column(Unicode(128), nullable=False)
    instructions = Column(UnicodeText, nullable=True)
    email = Column(Unicode(128), nullable=True)
    expires = Column(DateTime, nullable=False, default=func.now())
    winners = Column(Integer, nullable=False, default=1)
    max_entrants = Column(Integer, nullable=False, default=100)
    attributes = Column(MutableDict.as_mutable(JSONEncodedDict), nullable=True)

    # relationships
    entrants = relationship('Entrant', backref='contest', cascade='all,delete')


    def pick_winners(self):
        """
        Pick the winners of the contest.

        args:
            None
        returns:
            list - a list of winning names
        """

        results = []
        nwinners = self.winners
        if len(self.entrants) == 0:
            return []

        if len(self.entrants) < self.winners:
            nwinners = len(self.entrants)

        choices = [e.name for e in self.entrants]
        while len(results) < nwinners:
            name = secrets.choice(choices)
            if name not in results:
                results.append(name)

        return results


class Entrant(Base):
    """
    Entrant class definition
    """

    __tablename__ = 'entrants'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode, nullable=False, unique=False)
    contest_id = Column(Integer, ForeignKey('contests.id'))

    # ensure only one name sign-up per contest.
    __table_args__ = (
        UniqueConstraint('name', 'contest_id', name='_name_id_uc'),)


class Fossil(Base):
    """
    Fossil class definition. This is used only to keep a bare bones crumb
    of info about expired contests
    """

    __tablename__ = 'fossils'

    id = Column(Integer, primary_key=True)
    name = Column(Unicode(48), nullable=False, unique=True)
    expired = Column(DateTime, nullable=False, default=func.now())
    attributes = Column(MutableDict.as_mutable(JSONEncodedDict), nullable=True)



import string
import secrets
import datetime
import markdown
from tcw.apps.contest.models import Contest, Fossil
from .database import session
from .exc import ContestNotFound, FossilNotFound


def contest_by_name(name):
    """
    Get contest object by name
    """

    contest = session.query(Contest).filter(Contest.name == name).one()
    if not contest:
        raise ContestNotFound(f"No contest with name {name}")

    return contest


def fossil_by_name(name):
    """
    Get fossil object by name
    """

    fossil = session.query(Fossil).filter(Fossil.name == name).one()
    if not fossil:
        raise FossilNotFound(f"No fossil with name {name}")

    return fossil


def random_name(length=24):
    """
    create random name for contests. a mix of alphanum chars.

    args:
        - int name length
    returns:
        - str
    """

    letters = string.ascii_letters
    digits = string.digits
    alphabet = letters + digits
    name = ''

    while len(name) < length:
        name += secrets.choice(alphabet)

    return name


def expires_time(hours=1):
    """
    get a datetime object x number of hours in the future.

    args:
        - float hours into the future
    returns:
        - datateime object, None on  error
    """

    now = datetime.datetime.utcnow().replace(second=0, microsecond=0)
    later = now + datetime.timedelta(hours=hours)

    return later


def md_to_html(txt):
    """
    Convert markdown text to HTML.

    args:
        - str (markdown text)
    returns:
        str (html text)
    """

    html = markdown.markdown(txt)
    return html



from flask_wtf import FlaskForm, RecaptchaField
from wtforms import (StringField, IntegerField, TextAreaField, SubmitField,
    SelectField)
from wtforms.validators import DataRequired
from wtforms.fields import EmailField


class ContestForm(FlaskForm):
    """
    Form definition class for new contest    
    """

    title = StringField('contest title', validators=[DataRequired()],)
    instructions = TextAreaField('sign-up instructions', validators=[DataRequired()])
    email = EmailField('contest owner email', validators=[DataRequired()])
    winners = IntegerField('number of winners (1-50)', validators=[DataRequired()])
    maximum = IntegerField('max number of entrants (1-500)', validators=[DataRequired()])
    hours = SelectField('contest expires after', choices=[
        ('1', '1 hour'),
        ('4', '4 hours'),
        ('12', '12 hours'),
        ('24', '1 day'),
        ('72', '3 days'),
        ('120', '5 days'),
        ],  validators=[DataRequired()])
    recaptcha = RecaptchaField()
    submit = SubmitField('submit')


class SignupForm(FlaskForm):
    """
    Form definition for user signup form
    """

    name = StringField('sign up!')
    submit = SubmitField('submit')
